using System;
using UnityEngine;
using UnityEngine.Events;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Threading;
using MVR;
using MeshVR;
using System.Linq;
using Battlehub.RTCommon;
using GPUTools.Physics.Scripts.Behaviours;

namespace Stopper {
	public class ShoeHelper : MVRScript {
		DAZCharacterSelector characterSelector;
		
		List<ConfigurableJoint> toeJoints = new List<ConfigurableJoint>();
		
		string selectedClothingUid;
		DAZClothingItem shoeClothingItem;
		
		Vector3 shoeColliderDimensions = new Vector3(0.01f, 0.01f, 0.15f);
		Vector3 shoeColliderCenter = new Vector3(0.0f, 0.0f, 0.075f);
		Vector3 shoeColliderRotation = new Vector3(20.0f, 0f, 0.0f);
		
		float shoeFootDriveTarget = -45f;
		float shoeToeDriveTarget = 45f;

		bool isFirstSync = true;
		float originalShoeFootDriveTarget = 0f;
		float originalShoeToeDriveTarget = 0f;
		

		DAZBone lFoot;
		DAZBone rFoot;

		JSONStorableBool adjustController;

		bool isClothingPlugin = false;

		BoxCollider soleCollider;

		GameObject lHeelPreview;
		GameObject rHeelPreview;
		GameObject solePreview;

		Dictionary<GameObject, BoxCollider> colliderPreviews = new Dictionary<GameObject, BoxCollider>();

		List<GameObject> createdObjects = new List<GameObject>();

		JSONStorableBool showColliders;
		
		public override void Init() {
			WalkAndGetParentItems();
			// We found a clothing item in our hierarchy. We're probably being used as a clothing plugin.
			// Otherwise something very weird has happened...			
			isClothingPlugin = shoeClothingItem != null;
	
			characterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
			
			var physicsModel = containingAtom.GetStorableByID("PhysicsModel");
			foreach(Rigidbody rb in physicsModel.GetComponentsInChildren<Rigidbody>()){
				if(rb.name.Contains("SmallToe") || rb.name.Contains("BigToe") ){
					// There may be multple joints, but this seems to get the right one
					toeJoints.Add(rb.GetComponent<ConfigurableJoint>());
				}
			}

			lFoot = containingAtom.GetStorableByID("lFoot") as DAZBone;
			rFoot = containingAtom.GetStorableByID("rFoot") as DAZBone;
			
			if(!isClothingPlugin) {
				JSONStorableStringChooser clothingChooser = new JSONStorableStringChooser(
						"High Heel Clothing", 
						containingAtom.GetComponentsInChildren<DAZClothingItem>().Select(c => c.uid).ToList(),
						null, 
						"High Heel Clothing", 
						(string uid) => {
							if(shoeClothingItem != null){
								DisconnectClothingItem(shoeClothingItem);
							}
							
							shoeClothingItem = characterSelector.GetClothingItem(uid);

							
							if(shoeClothingItem != null && !shoeClothingItem.name.Contains("Harli")) {
								SyncShoeSettings(shoeClothingItem);
								ConnectClothingItem(shoeClothingItem);
							} else {
								SuperController.LogMessage($"Could not find Clothing {selectedClothingUid}");	
							}
						});
						
				RegisterStringChooser(clothingChooser);
				var selectShoePopup = CreateScrollablePopup(clothingChooser, false);
				selectShoePopup.popup.onOpenPopupHandlers += () => {
					clothingChooser.choices = containingAtom.GetComponentsInChildren<DAZClothingItem>().Select(c => c.uid).ToList();
				};
			}

			showColliders = new JSONStorableBool("Show Colliders", false);
			showColliders.isStorable = false;
			RegisterBool(showColliders);
			var showCollidersToggle = CreateToggle(showColliders, false);
			showCollidersToggle.toggle.onValueChanged.AddListener((val) => {
				foreach(var entry in colliderPreviews) {
					entry.Key.active = val;
				}
			});

			adjustController = new JSONStorableBool("Adjust Controller Rotation", true);
			RegisterBool(adjustController);
			var adjustRotationToggle = CreateToggle(adjustController, false);
			adjustRotationToggle.toggle.onValueChanged.AddListener((val) => {
				if(val) {
					AdjustControllerRotation(shoeFootDriveTarget);
				} else {
					AdjustControllerRotation(0f);
				}
			});

			
			AddSlider("Heel Collider Length", 0.15f, 0f, 0.5f, (float val) => {
				shoeColliderDimensions.z = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
			
			AddSlider("Heel Collider Width", 0.01f, 0f, 0.1f, (float val) => {
				shoeColliderDimensions.x = val;
				shoeColliderDimensions.y = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			AddSlider("Heel Collider Y Offset", 0.00f, -0.2f, 0.2f, (float val) => {
				shoeColliderCenter.y = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);

			AddSlider("Heel Collider Z Offset", 0.075f, -0.2f, 0.2f, (float val) => {
				shoeColliderCenter.z = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
			
			AddSlider("Heel Collider Rotation", 20f, -90f, 90f, (float val) => {
				shoeColliderRotation.x = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
			
			AddSlider("Toe Rotation Target", shoeToeDriveTarget, -75f, 75f, (float val) => {
				shoeToeDriveTarget = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
			
			AddSlider("Foot Rotation Target", shoeFootDriveTarget, -65f, 65f, (float val) => {
				shoeFootDriveTarget = val;
				SyncShoeSettings(shoeClothingItem);
			}, false);
			
			AddSlider("Toe Stiffness", 10f, 0f, 2000f, (float val) => {
				foreach(var joint in toeJoints){
					var slerpDrive = joint.slerpDrive;
					slerpDrive.positionSpring = val;
					joint.slerpDrive = slerpDrive;
				}
			}, true);
			
			AddSlider("Toe Damper", 10f, 0f, 2000f, (float val) => {
				foreach(var joint in toeJoints){
					var slerpDrive = joint.slerpDrive;
					slerpDrive.positionDamper = val;
					joint.slerpDrive = slerpDrive;
				}
			}, true);
			
			AddSlider("Toe Max Force", 10f, 0f, 2000f, (float val) => {
				foreach(var joint in toeJoints){
					var slerpDrive = joint.slerpDrive;
					slerpDrive.maximumForce = val;
					joint.slerpDrive = slerpDrive;
				}
			}, true);
	
			lHeelPreview = SetUpColliderPreview(characterSelector.leftShoeCollider);
			rHeelPreview = SetUpColliderPreview(characterSelector.rightShoeCollider);

			if(isClothingPlugin) {
				OnEnable();
			}
		}

		private void WalkAndGetParentItems(){
			GameObject parent = this.gameObject;
			while(containingAtom == null || parent != containingAtom.gameObject){
				if(parent.transform.parent == null){
					return;
				}
				
				if(shoeClothingItem == null){
					shoeClothingItem = parent.GetComponent<DAZClothingItem>();
				}

				if(containingAtom == null) {
                	containingAtom = parent.GetComponent<Atom>();
				}
				parent = parent.transform.parent.gameObject;
			}
		}

		private GameObject SetUpColliderPreview(BoxCollider parentCollider) {
			GameObject preview = GameObject.CreatePrimitive(PrimitiveType.Cube);
			
			createdObjects.Add(preview);
			colliderPreviews[preview] = parentCollider;

			preview.active = showColliders.val;
			foreach(var collider in preview.GetComponentsInChildren<Collider>(true)){
				collider.enabled = false;	
				UnityEngine.Object.Destroy(collider);
			}
			preview.transform.parent = parentCollider.transform;
			preview.transform.position = parentCollider.transform.position;
			preview.transform.rotation = parentCollider.transform.rotation;
			preview.transform.localScale = parentCollider.size;
			
			foreach(Renderer renderer in preview.GetComponentsInChildren<Renderer>()){
				renderer.material = new Material(Shader.Find("Battlehub/RTGizmos/Handles"));

				renderer.material.SetFloat("_Offset", 1f);
				renderer.material.SetFloat("_MinAlpha", 1f);
				Color color = Color.green;
				color.a = 0.3f;
				renderer.material.color = color;
			}
			return preview;
		}

		public void Update() {
			if(this.showColliders != null && showColliders.val) {
				foreach(var entry in colliderPreviews) {
					entry.Key.transform.localScale = entry.Value.size;
					entry.Key.transform.localPosition = entry.Value.center;
				}
			}
		}
		
		private void SyncShoeSettings(DAZClothingItem shoeClothingItem){
			if(isFirstSync) {
				originalShoeFootDriveTarget = shoeClothingItem.driveXAngleTarget;
				originalShoeToeDriveTarget = shoeClothingItem.drive2XAngleTarget;
				isFirstSync = false;
			}

			shoeClothingItem.colliderTypeLeft = DAZClothingItem.ColliderType.Shoe;
			shoeClothingItem.colliderTypeRight = DAZClothingItem.ColliderType.Shoe;
			
			shoeClothingItem.driveXAngleTargetController1Type = DAZClothingItem.ControllerType.RightToe;
			shoeClothingItem.driveXAngleTargetController2Type = DAZClothingItem.ControllerType.LeftToe;
			
			shoeClothingItem.drive2XAngleTargetController1Type = DAZClothingItem.ControllerType.RightFoot;
			shoeClothingItem.drive2XAngleTargetController2Type = DAZClothingItem.ControllerType.LeftFoot;
			
			shoeClothingItem.colliderDimensions = shoeColliderDimensions;
			
			shoeClothingItem.colliderRightCenter = shoeColliderCenter;
			shoeClothingItem.colliderLeftCenter = shoeColliderCenter;
			
			shoeClothingItem.colliderRightRotation = shoeColliderRotation;
			shoeClothingItem.colliderLeftRotation = shoeColliderRotation;
			
			shoeClothingItem.driveXAngleTarget = shoeToeDriveTarget;
			shoeClothingItem.drive2XAngleTarget = shoeFootDriveTarget;
			
			RefreshDazClothingItem();

			if(adjustController.val){
				AdjustControllerRotation(shoeFootDriveTarget);
			}
		}

		private void RefreshDazClothingItem(){
			if(shoeClothingItem != null) {
				if (shoeClothingItem.colliderRight != null)
				{
					shoeClothingItem.colliderRight.gameObject.SetActive(true);
					shoeClothingItem.colliderRight.transform.localEulerAngles = shoeClothingItem.colliderRightRotation;
					shoeClothingItem.colliderRight.size = shoeClothingItem.colliderDimensions;
					shoeClothingItem.colliderRight.center = shoeClothingItem.colliderRightCenter;
				}
				if (shoeClothingItem.colliderLeft != null)
				{
					shoeClothingItem.colliderLeft.gameObject.SetActive(true);
					shoeClothingItem.colliderLeft.transform.localEulerAngles = shoeClothingItem.colliderLeftRotation;
					shoeClothingItem.colliderLeft.size = shoeClothingItem.colliderDimensions;
					shoeClothingItem.colliderLeft.center = shoeClothingItem.colliderLeftCenter;
				}
				if (shoeClothingItem.driveXAngleTargetController1 != null)
				{
					shoeClothingItem.driveXAngleTargetController1.jointRotationDriveXTarget = shoeClothingItem.driveXAngleTarget;
				}
				if (shoeClothingItem.driveXAngleTargetController2 != null)
				{
					shoeClothingItem.driveXAngleTargetController2.jointRotationDriveXTarget = shoeClothingItem.driveXAngleTarget;
				}
				if (shoeClothingItem.drive2XAngleTargetController1 != null)
				{
					shoeClothingItem.drive2XAngleTargetController1.jointRotationDriveXTarget = shoeClothingItem.drive2XAngleTarget;
				}
				if (shoeClothingItem.drive2XAngleTargetController2 != null)
				{
					shoeClothingItem.drive2XAngleTargetController2.jointRotationDriveXTarget = shoeClothingItem.drive2XAngleTarget;
				}
			}
		}

		private void AdjustControllerRotation(float angle){
			foreach(var j in lFoot.GetComponents<ConfigurableJoint>()){
				var	controller = j.connectedBody.GetComponent<FreeControllerV3>();
				if(controller != null){
					var targetRotation = j.targetRotation.eulerAngles;
					targetRotation.x = angle;
					j.targetRotation = Quaternion.Euler(targetRotation);
				}
			}

			foreach(var j in rFoot.GetComponents<ConfigurableJoint>()){
				var	controller = j.connectedBody.GetComponent<FreeControllerV3>();
				if(controller != null){
					var targetRotation = j.targetRotation.eulerAngles;
					targetRotation.x = angle;
					j.targetRotation = Quaternion.Euler(targetRotation);
				}
			}
		}		

		
		// Copied from DAZCharacterSelector
		private void ConnectClothingItem(DAZClothingItem dci)
		{
			DAZClothingItem.ColliderType colliderTypeLeft = dci.colliderTypeLeft;
			if (colliderTypeLeft != DAZClothingItem.ColliderType.None)
			{
				if (colliderTypeLeft == DAZClothingItem.ColliderType.Shoe)
				{
					dci.colliderLeft = characterSelector.leftShoeCollider;
				}
			}
			else
			{
				dci.colliderLeft = null;
			}
			DAZClothingItem.ColliderType colliderTypeRight = dci.colliderTypeRight;
			if (colliderTypeRight != DAZClothingItem.ColliderType.None)
			{
				if (colliderTypeRight == DAZClothingItem.ColliderType.Shoe)
				{
					dci.colliderRight = characterSelector.rightShoeCollider;
				}
			}
			else
			{
				dci.colliderRight = null;
			}
			switch (dci.driveXAngleTargetController1Type)
			{
			case DAZClothingItem.ControllerType.None:
				dci.driveXAngleTargetController1 = null;
				break;
			case DAZClothingItem.ControllerType.LeftFoot:
				dci.driveXAngleTargetController1 = characterSelector.leftFootController;
				break;
			case DAZClothingItem.ControllerType.RightFoot:
				dci.driveXAngleTargetController1 = characterSelector.rightFootController;
				break;
			case DAZClothingItem.ControllerType.LeftToe:
				dci.driveXAngleTargetController1 = characterSelector.leftToeController;
				break;
			case DAZClothingItem.ControllerType.RightToe:
				dci.driveXAngleTargetController1 = characterSelector.rightToeController;
				break;
			}
			switch (dci.driveXAngleTargetController2Type)
			{
			case DAZClothingItem.ControllerType.None:
				dci.driveXAngleTargetController2 = null;
				break;
			case DAZClothingItem.ControllerType.LeftFoot:
				dci.driveXAngleTargetController2 = characterSelector.leftFootController;
				break;
			case DAZClothingItem.ControllerType.RightFoot:
				dci.driveXAngleTargetController2 = characterSelector.rightFootController;
				break;
			case DAZClothingItem.ControllerType.LeftToe:
				dci.driveXAngleTargetController2 = characterSelector.leftToeController;
				break;
			case DAZClothingItem.ControllerType.RightToe:
				dci.driveXAngleTargetController2 = characterSelector.rightToeController;
				break;
			}
			switch (dci.drive2XAngleTargetController1Type)
			{
			case DAZClothingItem.ControllerType.None:
				dci.drive2XAngleTargetController1 = null;
				break;
			case DAZClothingItem.ControllerType.LeftFoot:
				dci.drive2XAngleTargetController1 = characterSelector.leftFootController;
				break;
			case DAZClothingItem.ControllerType.RightFoot:
				dci.drive2XAngleTargetController1 = characterSelector.rightFootController;
				break;
			case DAZClothingItem.ControllerType.LeftToe:
				dci.drive2XAngleTargetController1 = characterSelector.leftToeController;
				break;
			case DAZClothingItem.ControllerType.RightToe:
				dci.drive2XAngleTargetController1 = characterSelector.rightToeController;
				break;
			}
			switch (dci.drive2XAngleTargetController2Type)
			{
			case DAZClothingItem.ControllerType.None:
				dci.drive2XAngleTargetController2 = null;
				break;
			case DAZClothingItem.ControllerType.LeftFoot:
				dci.drive2XAngleTargetController2 = characterSelector.leftFootController;
				break;
			case DAZClothingItem.ControllerType.RightFoot:
				dci.drive2XAngleTargetController2 = characterSelector.rightFootController;
				break;
			case DAZClothingItem.ControllerType.LeftToe:
				dci.drive2XAngleTargetController2 = characterSelector.leftToeController;
				break;
			case DAZClothingItem.ControllerType.RightToe:
				dci.drive2XAngleTargetController2 = characterSelector.rightToeController;
				break;
			}
			RefreshDazClothingItem();
		}
		
		private void DisconnectClothingItem(DAZClothingItem shoeClothingItem){
			shoeClothingItem.driveXAngleTargetController1.jointRotationDriveXTarget = 0f;
			shoeClothingItem.driveXAngleTargetController2.jointRotationDriveXTarget = 0f;
			shoeClothingItem.drive2XAngleTargetController1.jointRotationDriveXTarget = 0f;
			shoeClothingItem.drive2XAngleTargetController2.jointRotationDriveXTarget = 0f;

			shoeClothingItem.colliderTypeLeft = DAZClothingItem.ColliderType.None;
			shoeClothingItem.colliderTypeRight = DAZClothingItem.ColliderType.None;
			
			shoeClothingItem.driveXAngleTargetController1Type = DAZClothingItem.ControllerType.None;
			shoeClothingItem.driveXAngleTargetController2Type = DAZClothingItem.ControllerType.None;
			
			shoeClothingItem.drive2XAngleTargetController1Type = DAZClothingItem.ControllerType.None;
			shoeClothingItem.drive2XAngleTargetController2Type = DAZClothingItem.ControllerType.None;

			ConnectClothingItem(shoeClothingItem);

			AdjustControllerRotation(0f);
			RefreshDazClothingItem();
		}
		
		public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, UnityAction<float> onChange = null, bool rightSide = false)
		{
			JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
			storableFloat.storeType = JSONStorableParam.StoreType.Full;
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			uiSlider.rangeAdjustEnabled = true;
			RegisterFloat(storableFloat);
			
			if(onChange != null){
				uiSlider.slider.onValueChanged.AddListener(onChange);
			}
			
			return storableFloat;
		}
		
		private void Start(){
			
		}
	
			
		private void OnDisable()
		{
			if(shoeClothingItem != null){
				DisconnectClothingItem(shoeClothingItem);
			}
		}
		
		private void OnEnable()
		{
			if(shoeClothingItem != null){
				SyncShoeSettings(shoeClothingItem);
				ConnectClothingItem(shoeClothingItem);
			}
		}
		
		private void OnDestroy(){
			foreach(var go in createdObjects) {
				UnityEngine.Object.Destroy(go);
			}
		}
	}
}
